class Check:
    def __init__(self, combine):
        self.combine = combine

    def run(self):
        return None


class Message:
    def __init__(self, message, id, hint=None, object=None):
        self.message = message
        self.id = id
        self.hint = hint
        self.object = object

    def __str__(self):
        return (
            f"{self.id}: {self.message}\n  Hint: {self.hint}\n  Object: {self.object}"
        )

    def __repr__(self):
        return "<%s: message=%r, hint=%r, object=%r, id=%r>" % (
            self.__class__.__name__,
            self.message,
            self.hint,
            self.object,
            self.id,
        )

    @property
    def id_number(self):
        return self.id.split(".")[1]

    @property
    def color(self):
        if self.id_number.startswith("E"):
            return "red"
        elif self.id_number.startswith("W"):
            return "yellow"
        return None



import logging
import os
import time
from http.server import HTTPServer as BaseHTTPServer, SimpleHTTPRequestHandler
from fnmatch import fnmatch
import subprocess
import shlex

import click
from watchdog.observers import Observer
from watchdog.events import FileSystemEventHandler, FileCreatedEvent, DirModifiedEvent

from .files.template import TemplateFile


logger = logging.getLogger(__file__)


class Watcher:
    def __init__(self, path, combine):
        self.path = path
        self.combine = combine
        self.observer = Observer()
        self.event_handler = EventHandler(combine)

    def watch(self, while_running_func=None):
        self.observer.schedule(self.event_handler, self.path, recursive=True)
        self.observer.start()
        try:
            if while_running_func is not None:
                while_running_func()
            else:
                while True:
                    time.sleep(1)
        except (Exception, KeyboardInterrupt) as e:
            print(e)
            self.observer.stop()
        self.observer.join()


class EventHandler(FileSystemEventHandler):
    def __init__(self, combine, *args, **kwargs):
        self.combine = combine
        super().__init__(*args, **kwargs)

    def on_any_event(self, event):
        # if a file was moved or something, we only care about the destination
        event_path = event.dest_path if hasattr(event, "dest_path") else event.src_path

        if self.combine.is_in_output_path(event_path):
            # never need to process if in output path
            return

        # if matches a specific pattern, only use that
        for step in self.combine.config.steps:
            command = step["run"]
            for pattern in step["watch"]:
                # TODO remove ./ automatically?
                if fnmatch(event_path, pattern):
                    click.secho(
                        "Running command for matching config pattern", fg="cyan"
                    )
                    result = subprocess.run(shlex.split(command))
                    if result.returncode != 0:
                        click.secho(
                            "There was an error running a user command.", fg="red"
                        )
                    return

        if os.path.abspath(event_path) == os.path.abspath(self.combine.config_path):
            print(event)
            self.reload_combine()
            self.rebuild_site()

        if self.combine.is_in_content_paths(os.path.abspath(event_path)):
            print(event)
            if isinstance(event, (FileCreatedEvent, DirModifiedEvent)):
                self.reload_combine()

            file_obj = self.combine.get_file_obj_for_path(event_path)
            if isinstance(file_obj, TemplateFile):
                # if it was a template, we want to rebuild everything that uses it,
                # but right now we just rebuild the entire site
                self.rebuild_site()
            else:
                self.rebuild_site(only_paths=[os.path.abspath(event_path)])

    def reload_combine(self):
        click.secho("Reloading combine", fg="cyan")
        try:
            self.combine.reload()
        except Exception as e:
            logger.error("Error reloading", exc_info=e)
            click.secho("There was an error! See output above.", fg="red")

    def rebuild_site(self, only_paths=None):
        click.secho(f'Rebuilding {only_paths or "site"}', fg="cyan")
        try:
            self.combine.build(only_paths)
            click.secho("Site built", fg="green")
        except Exception as e:
            logger.error("Error building", exc_info=e)
            click.secho("There was an error! See output above.", fg="red")


class Server:
    def __init__(self, path, port=8000):
        self.path = path
        self.port = port
        self.httpd = HTTPServer(self.path, ("", self.port))

    def serve(self):
        click.secho(f"Serving at http://127.0.0.1:{self.port}", fg="green")
        self.httpd.serve_forever()


class HTTPHandler(SimpleHTTPRequestHandler):
    """This handler uses server.base_path instead of always using os.getcwd()"""

    def translate_path(self, path):
        path = SimpleHTTPRequestHandler.translate_path(self, path)
        relpath = os.path.relpath(path, os.getcwd())
        fullpath = os.path.join(self.server.base_path, relpath)
        return fullpath


class HTTPServer(BaseHTTPServer):
    """The main server, you pass in base_path which is the path you want to serve requests from"""

    def __init__(self, base_path, server_address, RequestHandlerClass=HTTPHandler):
        self.base_path = base_path
        BaseHTTPServer.__init__(self, server_address, RequestHandlerClass)



import os
import glob

from bs4 import BeautifulSoup

from .core import Check, Message
from .registry import register


@register
class PageTitlesCheck(Check):
    def run(self):
        messages = []
        page_titles = {}

        ignored_titles = ["Redirecting..."]

        for path in glob.iglob(
            os.path.join(self.combine.output_path, "**", "*.html"), recursive=True
        ):
            with open(path, "r") as f:
                soup = BeautifulSoup(f, "html.parser")
                title = soup.title
                if title:
                    if title.string in ignored_titles:
                        continue

                    if title.string in page_titles:
                        page_titles[title.string].append(path)
                    else:
                        page_titles[title.string] = [path]
                else:
                    messages.append(
                        Message(
                            "Page title is missing",
                            "seo.E001",
                            hint="Add a <title> in <head>",
                            object=path,
                        )
                    )

        for title, paths in page_titles.items():
            if len(paths) > 1:
                messages.append(
                    Message(
                        "Duplicate page title",
                        "seo.W001",
                        hint="Change the <title> attribute",
                        object=paths,
                    )
                )

            if len(title) > 60:
                messages.append(
                    Message(
                        "Long page title",
                        "seo.W002",
                        hint="Make the title shorter. https://moz.com/learn/seo/title-tag",
                        object=paths,
                    )
                )

        # TODO put brand at the end? Check if lots of titles start with the same thing and warn to move it to the end...

        return messages


@register
class AltTextCheck(Check):
    def run(self):
        messages = []

        for path in glob.iglob(
            os.path.join(self.combine.output_path, "**", "*.html"), recursive=True
        ):
            with open(path, "r") as f:
                soup = BeautifulSoup(f, "html.parser")
                for img in soup.findAll("img"):
                    if img.attrs.get("alt", None) is None:
                        messages.append(
                            Message(
                                "Alt text missing on img",
                                "seo.E002",
                                hint=f"Add alt attribute to img tag",
                                object=(img, path),
                            )
                        )

        return messages



import os

from .core import Check, Message
from .registry import register


@register
class EmptyBuildCheck(Check):
    """Ensures that the output directory actually has files in it"""

    def run(self):
        if not os.listdir(self.combine.output_path):
            return [Message("Output directory is empty", "empty.E001")]
